--
--
-- 30log goes here, for safety's sake.
--
--

local class = require 'lib.30log'

local SimpleWorldBase = class('WorldBase', {
	systems = {},
	renderSystems = {},
	entities = {},
	totalUpdates = 0,
	timeSpentUpdating = 0
})

function SimpleWorldBase:update(dt)
	self.totalUpdates = self.totalUpdates + 1
	self.timeSpentUpdating = self.timeSpentUpdating + dt
	
	for _, system in pairs(self.systems) do 
		system:_update(dt)
	end
	
	for id, entity in pairs(self.entities) do
		entity:update(dt)
		
		for _, system in pairs(self.systems) do
			if(system.process) then
				if(system:filter(entity)) then
					system:process(entity, dt)
				end
			end
		end	
	end
end

function SimpleWorldBase:addEntity(entity, name)
	
	if(name == nil) then
		name = string.random(20)
	end
		
	entity._id = name
	self.entities[name] = entity
end

function SimpleWorldBase:addSystem(system)
	table.insert(self.systems, system)
end

function SimpleWorldBase:addRenderSystem(system)
	table.insert(self.renderSystems, system)
end

function SimpleWorldBase:draw(dt)
	for _, system in pairs(self.systems) do 
		system:startDraw(dt)
	end
	
	for id, entity in pairs(self.entities) do
		entity:draw(dt)
		
		for _, system in pairs(self.systems) do
			if(system.draw) then
				if(system:filter(entity)) then
					system:draw(entity, dt)
				end
			end
		end	
	end
end

function SimpleWorldBase:remove(entity)
	if(entity._id == nil) then
		print('Attempted to remove entity not in system.')
	else
		self.entities[entity._id] = nil
	end
	
	return entity
end

local SimpleSystemBase = class('System', {
	totalUpdates = 0
})

function SimpleSystemBase:filter(entity)
	return true
end

function SimpleSystemBase:_update(dt)
	self.totalUpdates = self.totalUpdates + 1
	
	if(self.update) then
		self:update(dt)
	end
end

function SimpleSystemBase:startDraw(dt)
	--
	-- Do nothing.
	--
end

function SimpleSystemBase:draw(entity, dt)
	
end

local Entity = class('Entity', {
	x = 1, y = 1, z = 1,
	
	--
	-- Do some hitbox stuff.
	--
	velocity = { 
		x = 0,
		y = 0,
		max = {
			x = 0,
			y = 0
		},
		min = {
			x = 0,
			y = 0
		}
	},
	
	drag = {
		x = 1,
		y = 1
	},
	
	hitbox = {
		offset = {
			x = 0, y = 0
		},
		width = 0,
		height = 0
	},
	
	mode = {
		hovered = false,
		clicked = false
	},
	
	canInteract = false,
	
	behaviorList = {},
	componentList = {},
	componentOptions = {},
	behaviors = {}
})

function Entity:init(dt)
	if(#self.behaviorList > 0) then
		for name, options in pairs(self.behaviorList) do
			if(type(options) == 'string') then
				self:add(options)
			else
				self:add(name, options)
			end
		end
	end
	
	if(#self.componentList > 0) then
		for name, options in pairs(self.componentList) do
			self[options] = Game.ecs.component(options, self)
			
			if(self.componentOptions[options]) then
				for key, value in pairs(self.componentOptions[options]) do 
					self[options][key] = value
				end
			end
		end
	end
end

function Entity:update(dt)
	for name, behavior in pairs(self.behaviors) do
		if(not behavior._awake) then
			behavior:wake(dt)
			behavior._awake = true
		end
	
		behavior:update(dt)
	end
	
	if(contains(Game.mouse.x, Game.mouse.y, self.x + self.hitbox.offset.x, self.y + self.hitbox.offset.y, self.x + self.hitbox.offset.x + self.hitbox.width, self.y + self.hitbox.offset.y + self.hitbox.height)) then
		for name, behavior in pairs(self.behaviors) do
			if(not self.mode.hovered) then
				behavior:onHover(dt)
				
				if(Game.debug.showBoxes) then
					love.graphics.rectangle("line", self.x + self.hitbox.offset.x, self.y + self.hitbox.offset.y, self.x + self.hitbox.offset.x + self.hitbox.width, self.y + self.hitbox.offset.y + self.hitbox.height)
				end
				
				self.mode.hovered = true
			end
			
			if(Game.mouse.clicked and not self.mode.clicked) then
				behavior:onMouseDown(dt)			
				
				if(Game.mouse.leftClick) then
					behavior:onLeftClick(dt)
				end
					
				if(Game.mouse.rightClick) then
					behavior:onRightClick(dt)
				end
			end
		end
		
		if(Game.mouse.clicked) then
			self.mode.clicked = true
		else
			self.mode.clicked = false
		end
	else
		for name, behavior in pairs(self.behaviors) do
			if(self.mode.hovered) then
				behavior:onLeave(dt)
				self.mode.hovered = false
			end
		end
	end	
end

function Entity:draw(dt)
	for _, behavior in pairs(self.behaviors) do
		if(behavior.draw and behavior._awake) then
			behavior:draw(dt)
		end
	end
end

function Entity:add(behavior, options)
	assert((self.class ~= nil), 'Attached behavior to class. Attach a behavior to entities instead.')
	
	local _behavior = 0
	
	if(type(behavior) == 'string') then
		_behavior = require(string.format('game.behaviors.%s', behavior)):new(self)	
	else
		_behavior = behavior
	end
	
	if(options ~= nil) then
		for key, value in pairs(options) do
			_behavior[key] = value
		end
	end
	
	self.behaviors[_behavior.name] = _behavior
end

function Entity:remove(behavior)
	if(type(behavior) == 'string') then
		self.behaviors[behavior] = nil
	else
		self.behaviors[behavior.name] = nil
	end
end

function Entity:get(behavior)
	if(type(behavior) ~= 'string') then
		return false
	else
		return self.behaviors[behavior]
	end
end

local Behavior = class('Behavior', {
	_awake = false
})

function Behavior:init(parent)
	if(parent ~= nil) then
		self.parent = parent
	end
end

function Behavior:update(dt)
	return true
end

function Behavior:wake(dt)
	return true
end

function Behavior:onMouseDown(dt)
	return true
end

function Behavior:onMouseUp(dt)
	return true
end

function Behavior:onLeftClick(dt)
	return true
end

function Behavior:onRightClick(dt)
	return true
end

function Behavior:onMouseWheelUp(dt)
	return true
end

function Behavior:onMouseWheelDown(dt)
	return true
end

function Behavior:onHover(dt)
	return true
end

function Behavior:onLeave(dt)
	return true
end

function Behavior:update(dt)
	return true
end

function Behavior:draw(dt)
	return true
end

local EntityTracker = class('EntityTracker', {
	get = function(self, entity)
		return Game.worlds.scenes[Game.worlds.activeScene].entities[entity]
	end,
	
	remove = function(self, entity)
		if(type(entity) == 'string') then
			local entityID = Game.world.scenes[Game.worlds.activeScene].entities[entity]
			
			if(entityID == nil) then
				return false
			else
				if(entityID.Hitbox) then
					Game.physics:remove(entityID.Hitbox.shape)
				end
			end
			
			Game.worlds.scenes[Game.worlds.activeScene].entities[entity]= nil
		else
			if(entity.Hitbox) then
				Game.physics:remove(entity.Hitbox.shape)
				entity.Hitbox = nil
			end
			
			Game.worlds.scenes[Game.worlds.activeScene].entities[entity._id] = nil
		end
		
		return entity
	end,
	
	add = function(self, entity, name)
		if(name ~= nil and entity._id == nil) then
			entity._id = name
		end
		
		if(entity._id ~= nil and name == nil) then
		 	name = entity._id
		end
		
		if(entity._id == nil and name == nil) then
			name = string.random(20)
		end
		
		Game.worlds.scenes[Game.worlds.activeScene]:addEntity(entity, name)
	end
})

return class('ECS', {
	_component = class('Component'),
	component = function(name, entity)
		if(name ~= nil) then
			if(Game.ecs.components[name] == nil) then
				Game.ecs.components[name] = require('game.components.' .. name)
			end
				
			return Game.ecs.components[name]:new()
		else
			return Game.ecs._component
		end
	end,
	components = {},
	behavior = Behavior,
	entity = Entity,
	entities = EntityTracker:new(),
	system = SimpleSystemBase,
	world = SimpleWorldBase
}):new()