--
--
-- The main.
--
--

Game = {
	debug = {
		showStats = false,
		showBoxes = false,
		timeRunning = 0
	},
	
	cargo = require 'lib.cargo',
	class = require 'lib.30log',
	
	colorLayer = {
		r = 255, g = 255, b = 255
	},
	
	ecs = require 'lib.simple-ecs',
	file = require 'lib.ini_parser',
	
	fonts = {
		game = 0,
		debug = 0,
	},
	
	layerOrder = {
		'background', 'midground', 'foreground', 'interface', 'debug'
	},
	
	mouse = {
		x = 1,
		y = 1,
		button_1 = false,
		button_2 = false,
		mouseWheelUp = false,
		mouseWheelDown = false
	},
	
	particles = {
		
	},
	
	physics = 0,
	
	settings = {
		volume = {
			master = 1,
			music = 1,
			sfx = 1
		},
		colorLayer = {
			r = 255,
			g = 255,
			b = 255
		}
	},
	
	sound = require 'lib.sound',
	
	state = {
		objects = {},
		
		inputState = 0,
		lastClick = 0,
		clickState = 0,
		clicks = 0,
		clickValue = 1,
		gravity = 300,
		floor_y = 400,
		setting = 'basic'
	}
}

function love.load()
	print('Loading assets.')
	require 'game.functions'

	Game.physics = (require 'lib.collider')(100, onCollide)
	Game.physics:resetHash()
	
	love.filesystem.setIdentity('oh_my_boy')
	
	Game.assets = Game.cargo.init{
		dir = '/assets',
		loaders = {
			ini = function(file) 				
				return Game.file.load(file)
			end,
			mp3 = love.sound.newSoundData,
			wav = love.sound.newSoundData
		}
	}
	
	if(Game.state.clicks < 0) then
		Game.state.clicks = 0
	end
	
	Game.strings = Game.assets.text.strings
	Game.definitions = Game.assets.text.definitions
	
	Game.settings = table.merge(Game.settings, Game.assets.text.settings)
	
	if(love.filesystem.exists('settings.ini')) then
		Game.settings = table.merge(Game.settings, Game.file.load('settings.ini'))
	end
	
	Game.state.status = Game.strings.Status.EggStart
	
	Game.fonts.game = Game.assets.fonts.font(32)
	Game.fonts.game:setFilter("nearest", "nearest")
	
	print('Creating worlds.')
	Game.worlds = require 'game.scene'
	
	love.window.setTitle(Game.strings.Game.Title)
	love.window.setMode(800, 600)
	love.window.setIcon(Game.assets.images.icon:getData())
	love.graphics.setDefaultFilter("nearest", "nearest")
	
	Game.layers = {}
	
	for _, name in pairs(Game.layerOrder) do 
		Game.layers[name] = love.graphics.newCanvas(800, 600)
	end
	
	Game.particles = love.graphics.newParticleSystem(Game.assets.images.particle, 5000)
	Game.particles:setLinearDamping(1)
	Game.particles:setLinearAcceleration(-25, -100, 25, 10)
	Game.particles:setRadialAcceleration(20, 50)
	Game.particles:setParticleLifetime(0.5, 0.75)
	Game.particles:setSizeVariation(0)
	Game.particles:setSizes(10, 5, 2.5, 1)
	Game.particles:setSpinVariation(1)
	Game.particles:setEmissionRate(0)
	Game.particles:start()
end

function love.update(dt)
	(require 'lib.lovebird').update()
	Game.sound:volume("background", Game.settings.volume.master * Game.settings.volume.music)
	Game.sound:volume("sfx", Game.settings.volume.master * Game.settings.volume.sfx)
	
	Game.debug.timeRunning = Game.debug.timeRunning + dt
	
	Game.particles:update(dt)
	
	Game.mouse = {
		cursor = 'arrow',
		x = love.mouse.getX(),
		y = love.mouse.getY(),
		leftClick = love.mouse.isDown(1),
		rightClick = love.mouse.isDown(2),
		middleClick = love.mouse.isDown(3),
		mouseWheelUp = false,
		mouseWheelDown = false
	}
	
	Game.mouse.clicked = (Game.mouse.leftClick or Game.mouse.rightClick or Game.mouse.middleClick)
	
	if(Game.mouse.clicked) then
		Game.state.lastClick = 0
	else
		Game.state.lastClick = Game.state.lastClick + dt	
	end
	
	Game.worlds:update(dt)
	if(Game.state.clicks < 0) then
		Game.state.clicks = 0
	end
end

function love.draw(dt)
	for name, canvas in pairs(Game.layers) do
		love.graphics.setCanvas(canvas)
		love.graphics.clear()
	end
	
	love.graphics.setCanvas()
	
	love.graphics.setFont(Game.fonts.game)
	
	for color, value in pairs(Game.settings.colorLayer) do
		Game.settings.colorLayer[color] = math.clamp(value, 0, 255, true)
	end
	
	Game.worlds:draw(dt)
	love.graphics.setCanvas()
	
	for _, name in pairs(Game.layerOrder) do
		love.graphics.draw(Game.layers[name])
	end
	
	local lastColor = love.graphics.getColor()

	love.graphics.draw(Game.particles, 1, 1)
	
	love.graphics.setBlendMode('multiply')
	love.graphics.setColor(Game.settings.colorLayer.r, Game.settings.colorLayer.g, Game.settings.colorLayer.b)
	love.graphics.rectangle("fill", 0, 0, love.graphics.getWidth(), love.graphics.getHeight())
	
	love.graphics.reset()
	
	if(Game.debug.showStats) then
		love.graphics.print('Debug vision - version 0.0.1a', 0, 0)
		love.graphics.print("FPS: " .. math.ceil(1 / love.timer.getAverageDelta()) .. " Delta: " .. love.timer.getDelta(), 0, 40)
		love.graphics.print("R: " .. Game.settings.colorLayer.r .. ", G: " .. Game.settings.colorLayer.g .. ", B: " .. Game.settings.colorLayer.b, 0, 28)
		love.graphics.print(string.format("Mouse X: %i, Y: %i", Game.mouse.x, Game.mouse.y), 0, 58)
		love.graphics.print(string.format("Time running: %i", math.floor(Game.debug.timeRunning)), 0, 70)
	end
end

function love.keypressed(key, scancode, isrepeat)
	if(key == 'f1') then
		Game.debug.showStats = not Game.debug.showStats
		Game.debug.showBoxes = not Game.debug.showBoxes
	elseif(key == 'f12') then
		--
		-- F12 screencaps.
		--
		local screenshot = love.graphics.newScreenshot();
		if(not love.filesystem.isDirectory('screencaps')) then
			love.filesystem.createDirectory('screencaps')
		end
		
		local fileData = screenshot:encode('png', 'screencaps/' .. os.time() .. '.png');
		print('Screenshot taken: ' .. fileData:getFilename())
	end
end